@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003 Choe Hwanjin
+ * Copyright (C) 2003,2004 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <gtk/gtk.h>
+#include <gdk/gdkx.h>
 #include <glib.h>
 
 #include "gettext.h"
@@ -121,7 +122,7 @@ nabi_ic_init_values(NabiIC *ic)
     ic->mode = NABI_INPUT_MODE_DIRECT;
 
     /* preedit attr */
-    ic->preedit.window = 0;
+    ic->preedit.window = NULL;
     ic->preedit.width = 1;	/* minimum window size is 1 x 1 */
     ic->preedit.height = 1;	/* minimum window size is 1 x 1 */
     ic->preedit.area.x = 0;
@@ -135,9 +136,9 @@ nabi_ic_init_values(NabiIC *ic)
     ic->preedit.spot.x = 0;
     ic->preedit.spot.y = 0;
     ic->preedit.cmap = 0;
-    ic->preedit.gc = 0;
-    ic->preedit.foreground = nabi_server->preedit_fg;
-    ic->preedit.background = nabi_server->preedit_bg;
+    ic->preedit.gc = NULL;
+    ic->preedit.foreground = nabi_server->preedit_fg.pixel;
+    ic->preedit.background = nabi_server->preedit_bg.pixel;
     ic->preedit.bg_pixmap = 0;
     ic->preedit.cursor = 0;
     ic->preedit.base_font = NULL;
@@ -273,9 +274,9 @@ nabi_ic_destroy(NabiIC *ic)
     ic->preedit.start = False;
 
     /* destroy preedit window */
-    if (ic->preedit.window != 0) {
-	XDestroyWindow(nabi_server->display, ic->preedit.window);
-	ic->preedit.window = 0;
+    if (ic->preedit.window != NULL) {
+	gdk_window_destroy(ic->preedit.window);
+	ic->preedit.window = NULL;
     }
 
     /* destroy fontset data */
@@ -331,18 +332,41 @@ nabi_ic_real_destroy(NabiIC *ic)
     nabi_free(ic->preedit.base_font);
     nabi_free(ic->status_attr.base_font);
 
-    if (ic->preedit.window != 0)
-	XDestroyWindow(nabi_server->display, ic->preedit.window);
+    if (ic->preedit.window != NULL)
+	gdk_window_destroy(ic->preedit.window);
 
     if (ic->preedit.font_set)
 	nabi_fontset_free(nabi_server->display, ic->preedit.font_set);
 
-    if (ic->preedit.gc)
-	XFreeGC(nabi_server->display, ic->preedit.gc);
+    if (ic->preedit.gc != NULL)
+	g_object_unref(G_OBJECT(ic->preedit.gc));
 
     nabi_free(ic);
 }
 
+static void
+nabi_ic_preedit_gdk_draw_string(NabiIC *ic, char *str, int size)
+{
+    GdkGC *gc;
+    PangoLayout *layout;
+    int width = 12, height = 12;
+
+    if (ic->preedit.window == NULL)
+	return;
+
+    if (ic->preedit.gc == NULL)
+	gc = nabi_server->gc;
+    else
+	gc = ic->preedit.gc;
+
+    layout = gtk_widget_create_pango_layout(nabi_server->widget, str);
+    pango_layout_get_pixel_size(layout, &width, &height);
+    gdk_window_resize(ic->preedit.window, width + 3, height + 3);
+    gdk_window_clear(ic->preedit.window);
+    gdk_draw_layout(ic->preedit.window, gc, 2, 2, layout);
+    g_object_unref(G_OBJECT(layout));
+}
+
 static void
 nabi_ic_preedit_draw_string(NabiIC *ic, char *str, int size)
 {
@@ -353,17 +377,17 @@ nabi_ic_preedit_draw_string(NabiIC *ic, char *str, int size)
 	return;
     if (ic->preedit.font_set == 0)
 	return;
-    if (ic->preedit.gc != 0)
-	gc = ic->preedit.gc;
+    if (ic->preedit.gc != NULL)
+	gc = gdk_x11_gc_get_xgc(ic->preedit.gc);
     else 
-	gc = nabi_server->gc;
+	gc = gdk_x11_gc_get_xgc(nabi_server->gc);
 
     width = Xutf8TextEscapement(ic->preedit.font_set, str, size);
     if (ic->preedit.width != width) {
 	ic->preedit.width = width;
 	ic->preedit.height = ic->preedit.ascent + ic->preedit.descent;
-	XResizeWindow(nabi_server->display, ic->preedit.window,
-		      ic->preedit.width, ic->preedit.height);
+	gdk_window_resize(ic->preedit.window,
+		          ic->preedit.width, ic->preedit.height);
     }
 
     /* if preedit window is out of focus window 
@@ -371,13 +395,13 @@ nabi_ic_preedit_draw_string(NabiIC *ic, char *str, int size)
     if (ic->preedit.area.width != 0) {
 	if (ic->preedit.spot.x + ic->preedit.width > ic->preedit.area.width) {
 	    ic->preedit.spot.x = ic->preedit.area.width - ic->preedit.width;
-	    XMoveWindow(nabi_server->display, ic->preedit.window,
-			ic->preedit.spot.x, 
-			ic->preedit.spot.y - ic->preedit.ascent);
+	    gdk_window_move(ic->preedit.window,
+			    ic->preedit.spot.x, 
+			    ic->preedit.spot.y - ic->preedit.ascent);
 	}
     }
     Xutf8DrawImageString(nabi_server->display,
-		         ic->preedit.window,
+		         GDK_WINDOW_XWINDOW(ic->preedit.window),
 		         ic->preedit.font_set,
 		         gc,
 		         0,
@@ -395,43 +419,63 @@ nabi_ic_preedit_draw(NabiIC *ic)
 	return;
 
     nabi_ic_get_preedit_string(ic, buf, sizeof(buf), &len, &size);
-    nabi_ic_preedit_draw_string(ic, buf, size);
+    if (ic->input_style & XIMPreeditPosition) {
+	nabi_ic_preedit_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditArea) {
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
+    }
 }
 
 /* map preedit window */
 static void
 nabi_ic_preedit_show(NabiIC *ic)
 {
-    if (ic->preedit.window == 0)
+    if (ic->preedit.window == NULL)
 	return;
 
     /* draw preedit only when ic have any hangul data */
     if (!nabi_ic_is_empty(ic))
-	XMapWindow(nabi_server->display, ic->preedit.window);
+	gdk_window_show(ic->preedit.window);
 }
 
 /* unmap preedit window */
 static void
 nabi_ic_preedit_hide(NabiIC *ic)
 {
-    if (ic->preedit.window == 0)
+    if (ic->preedit.window == NULL)
 	return;
 
-    XUnmapWindow(nabi_server->display, ic->preedit.window);
+    gdk_window_hide(ic->preedit.window);
 }
 
 /* move and resize preedit window */
 static void
 nabi_ic_preedit_configure(NabiIC *ic)
 {
-    if (ic->preedit.window == 0)
+    if (ic->preedit.window == NULL)
 	return;
 
-    XMoveResizeWindow(nabi_server->display, ic->preedit.window,
-		      ic->preedit.spot.x + 1, 
-		      ic->preedit.spot.y - ic->preedit.ascent, 
-		      ic->preedit.width,
-		      ic->preedit.height);
+    if (ic->input_style & XIMPreeditPosition) {
+	gdk_window_move_resize(ic->preedit.window,
+			       ic->preedit.spot.x + 1, 
+			       ic->preedit.spot.y - ic->preedit.ascent, 
+			       ic->preedit.width,
+			       ic->preedit.height);
+    } else if (ic->input_style & XIMPreeditArea) {
+	gdk_window_move_resize(ic->preedit.window,
+			       ic->preedit.area.x + 1, 
+			       ic->preedit.area.y, 
+			       ic->preedit.width,
+			       ic->preedit.height);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	gdk_window_move_resize(ic->preedit.window,
+			       ic->preedit.spot.x + 1, 
+			       ic->preedit.spot.y - ic->preedit.ascent, 
+			       ic->preedit.width,
+			       ic->preedit.height);
+    }
 }
 
 static GdkFilterReturn
@@ -443,17 +487,17 @@ gdk_event_filter(GdkXEvent *xevent, GdkEvent *gevent, gpointer data)
     if (ic == NULL)
 	return GDK_FILTER_REMOVE;
 
-    if (ic->preedit.window == 0)
+    if (ic->preedit.window == NULL)
 	return GDK_FILTER_REMOVE;
 
-    if (event->xany.window != ic->preedit.window)
+    if (event->xany.window != GDK_WINDOW_XWINDOW(ic->preedit.window))
 	return GDK_FILTER_CONTINUE;
 
     switch (event->type) {
     case DestroyNotify:
 	if (!nabi_ic_is_destroyed(ic)) {
 	    /* preedit window is destroyed, so we set it 0 */
-	    ic->preedit.window = 0;
+	    ic->preedit.window = NULL;
 	    //nabi_ic_destroy(ic);
 	}
 	return GDK_FILTER_REMOVE;
@@ -473,45 +517,41 @@ gdk_event_filter(GdkXEvent *xevent, GdkEvent *gevent, gpointer data)
 static void
 nabi_ic_preedit_window_new(NabiIC *ic)
 {
-    Window parent;
-    int cmask = 0;
-    XSetWindowAttributes attributes;
-    GdkWindow *gdk_window;
+    GdkWindow *parent = NULL;
+    GdkWindowAttr attr;
+    gint mask;
 
     if (ic->focus_window != 0)
-	parent = ic->focus_window;
+	parent = gdk_window_foreign_new(ic->focus_window);
     else if (ic->client_window != 0)
-	parent = ic->client_window;
+	parent = gdk_window_foreign_new(ic->client_window);
     else
 	return;
 
-    ic->preedit.window = XCreateSimpleWindow(nabi_server->display,
-					     parent,
-					     ic->preedit.spot.x,
-					     ic->preedit.spot.y
-					      - ic->preedit.ascent,
-					     ic->preedit.width,
-					     ic->preedit.height,
-					     0,
-					     nabi_server->preedit_fg,
-					     nabi_server->preedit_bg);
-
+    attr.wclass = GDK_INPUT_OUTPUT;
+    attr.event_mask = GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK;
+    attr.window_type = GDK_WINDOW_TEMP;
+    attr.x = ic->preedit.spot.x;
+    attr.y = ic->preedit.spot.y - ic->preedit.ascent;
+    attr.width = ic->preedit.width;
+    attr.height = ic->preedit.height;
     /* set override-redirect to true
      * we should set this to show preedit window on qt apps */
-    cmask = CWOverrideRedirect;
-    attributes.override_redirect = True;
-    XChangeWindowAttributes(nabi_server->display, ic->preedit.window,
-			    cmask, &attributes);
+    attr.override_redirect = TRUE;
+    mask = GDK_WA_X | GDK_WA_Y | GDK_WA_NOREDIR;
 
-    XSelectInput(nabi_server->display, ic->preedit.window,
-		 ExposureMask | StructureNotifyMask);
+    ic->preedit.window = gdk_window_new(parent, &attr, mask);
+    gdk_window_set_background(ic->preedit.window, &(nabi_server->preedit_bg));
+
+    if (ic->preedit.gc != NULL) {
+	ic->preedit.gc = gdk_gc_new(ic->preedit.window);
+	gdk_gc_set_foreground(ic->preedit.gc, &(nabi_server->preedit_fg));
+	gdk_gc_set_background(ic->preedit.gc, &(nabi_server->preedit_bg));
+    }
 
     /* install our preedit window event filter */
-    gdk_window = gdk_window_foreign_new(ic->preedit.window);
-    if (gdk_window != NULL) {
-	gdk_window_add_filter(gdk_window, gdk_event_filter, (gpointer)ic);
-	g_object_unref(gdk_window);
-    }
+    gdk_window_add_filter(ic->preedit.window, gdk_event_filter, (gpointer)ic);
+    g_object_unref(G_OBJECT(parent));
 }
 
 static void
@@ -519,49 +559,48 @@ nabi_ic_set_focus_window(NabiIC *ic, Window focus_window)
 {
     ic->focus_window = focus_window;
 
-    if (!(ic->input_style & XIMPreeditPosition))
-	return;
-
-    /* For Preedit Position aka Over the Spot */
-    if (ic->preedit.window == 0)
-	nabi_ic_preedit_window_new(ic);
+    if (ic->input_style & XIMPreeditPosition) {
+	if (ic->preedit.window == NULL)
+	    nabi_ic_preedit_window_new(ic);
+    } if (ic->input_style & XIMPreeditArea) {
+	if (ic->preedit.window == NULL)
+	    nabi_ic_preedit_window_new(ic);
+    } if (ic->input_style & XIMPreeditNothing) {
+	if (ic->preedit.window == NULL)
+	    nabi_ic_preedit_window_new(ic);
+    }
 }
 
 static void
 nabi_ic_set_preedit_foreground(NabiIC *ic, unsigned long foreground)
 {
-    foreground = nabi_server->preedit_fg;
     ic->preedit.foreground = foreground;
 
     if (ic->focus_window == 0)
 	return;
 
-    if (ic->preedit.gc) {
-	XSetForeground(nabi_server->display, ic->preedit.gc, foreground);
+    if (ic->preedit.gc != NULL) {
+	gdk_gc_set_foreground(ic->preedit.gc, &(nabi_server->preedit_fg));
     } else {
-	XGCValues values;
-	values.foreground = foreground;
-	ic->preedit.gc = XCreateGC(nabi_server->display, nabi_server->window,
-		       GCForeground, &values);
+	ic->preedit.gc = gdk_gc_new(nabi_server->widget->window);
+	gdk_gc_set_foreground(ic->preedit.gc, &(nabi_server->preedit_fg));
     }
 }
 
 static void
 nabi_ic_set_preedit_background(NabiIC *ic, unsigned long background)
 {
-    background = nabi_server->preedit_bg;
+    GdkColor color = { background, 0, 0, 0 };
     ic->preedit.background = background;
 
     if (ic->focus_window == 0)
 	return;
 
-    if (ic->preedit.gc) {
-	XSetBackground(nabi_server->display, ic->preedit.gc, background);
+    if (ic->preedit.gc != NULL) {
+	gdk_gc_set_background(ic->preedit.gc, &color);
     } else {
-	XGCValues values;
-	values.background = background;
-	ic->preedit.gc = XCreateGC(nabi_server->display, nabi_server->window,
-		       GCBackground, &values);
+	ic->preedit.gc = gdk_gc_new(nabi_server->widget->window);
+	gdk_gc_set_background(ic->preedit.gc, &color);
     }
 }
 
@@ -615,6 +654,25 @@ nabi_ic_set_spot(NabiIC *ic, XPoint *point)
 	nabi_ic_preedit_show(ic);
 }
 
+static void
+nabi_ic_set_area(NabiIC *ic, XRectangle *rect)
+{
+    if (rect == NULL)
+	return;
+
+    ic->preedit.area.x = rect->x; 
+    ic->preedit.area.y = rect->y; 
+    ic->preedit.area.width = rect->width; 
+    ic->preedit.area.height = rect->height; 
+
+    nabi_ic_preedit_configure(ic);
+
+    if (nabi_ic_is_empty(ic))
+	nabi_ic_preedit_hide(ic);
+    else
+	nabi_ic_preedit_show(ic);
+}
+
 #define streql(x, y)	(strcmp((x), (y)) == 0)
 
 void
@@ -651,7 +709,7 @@ nabi_ic_set_values(NabiIC *ic, IMChangeICStruct *data)
 	    nabi_ic_set_preedit_background(ic,
 		    *(unsigned long*)preedit_attr->value);
 	} else if (streql(XNArea, preedit_attr->name)) {
-	    ic->preedit.area = *(XRectangle*)preedit_attr->value;
+	    nabi_ic_set_area(ic, (XRectangle*)preedit_attr->value);
 	} else if (streql(XNLineSpace, preedit_attr->name)) {
 	    ic->preedit.line_space = *(CARD32*)preedit_attr->value;
 	} else if (streql(XNPreeditState, preedit_attr->name)) {
@@ -699,17 +757,22 @@ nabi_ic_get_values(NabiIC *ic, IMChangeICStruct *data)
     for (i = 0; i < data->ic_attr_num; i++, ic_attr++) {
 	if (streql(XNFilterEvents, ic_attr->name)) {
 	    ic_attr->value = (void *)malloc(sizeof(CARD32));
-	    *(CARD32*)ic_attr->value
-		= KeyPressMask | KeyReleaseMask;
+	    *(CARD32*)ic_attr->value = KeyPressMask | KeyReleaseMask;
 	    ic_attr->value_length = sizeof(CARD32);
 	} else if (streql(XNInputStyle, ic_attr->name)) {
 	    ic_attr->value = (void *)malloc(sizeof(INT32));
-	    *(INT32*)ic_attr->value
-		= ic->input_style;
+	    *(INT32*)ic_attr->value = ic->input_style;
 	    ic_attr->value_length = sizeof(INT32);
 	} else if (streql(XNSeparatorofNestedList, ic_attr->name)) {
 	    /* FIXME: what do I do here? */
 	    ;
+	} else if (streql(XNPreeditState, ic_attr->name)) {
+	    /* some java applications need XNPreeditState attribute in
+	     * IC attribute instead of Preedit attributes
+	     * so we support XNPreeditState attr here */
+	    ic_attr->value = (void *)malloc(sizeof(XIMPreeditState));
+	    *(XIMPreeditState*)ic_attr->value = ic->preedit.state;
+	    ic_attr->value_length = sizeof(XIMPreeditState);
 	} else {
 	    fprintf(stderr, _("Nabi: get unknown ic attributes: %s\n"),
 		ic_attr->name);
@@ -718,42 +781,32 @@ nabi_ic_get_values(NabiIC *ic, IMChangeICStruct *data)
     
     for (i = 0; i < data->preedit_attr_num; i++, preedit_attr++) {
 	if (streql(XNArea, preedit_attr->name)) {
-	    preedit_attr->value 
-		= (void *)malloc(sizeof(XRectangle));
-	    *(XRectangle*)preedit_attr->value 
-		= ic->preedit.area;
+	    preedit_attr->value = (void *)malloc(sizeof(XRectangle));
+	    *(XRectangle*)preedit_attr->value = ic->preedit.area;
 	    preedit_attr->value_length = sizeof(XRectangle);
 	} else if (streql(XNAreaNeeded, preedit_attr->name)) {
-	    preedit_attr->value
-		= (void *)malloc(sizeof(XRectangle));
-	    *(XRectangle*)preedit_attr->value
-		= ic->preedit.area_needed;
+	    preedit_attr->value = (void *)malloc(sizeof(XRectangle));
+	    *(XRectangle*)preedit_attr->value = ic->preedit.area_needed;
 	    preedit_attr->value_length = sizeof(XRectangle);
 	} else if (streql(XNSpotLocation, preedit_attr->name)) {
 	    preedit_attr->value = (void *)malloc(sizeof(XPoint));
-	    *(XPoint*)preedit_attr->value 
-		= ic->preedit.spot;
+	    *(XPoint*)preedit_attr->value = ic->preedit.spot;
 	    preedit_attr->value_length = sizeof(XPoint);
 	} else if (streql(XNForeground, preedit_attr->name)) {
 	    preedit_attr->value = (void *)malloc(sizeof(long));
-	    *(long*)preedit_attr->value
-		= ic->preedit.foreground;
+	    *(long*)preedit_attr->value = ic->preedit.foreground;
 	    preedit_attr->value_length = sizeof(long);
 	} else if (streql(XNBackground, preedit_attr->name)) {
 	    preedit_attr->value = (void *)malloc(sizeof(long));
-	    *(long*)preedit_attr->value
-		= ic->preedit.background;
+	    *(long*)preedit_attr->value = ic->preedit.background;
 	    preedit_attr->value_length = sizeof(long);
 	} else if (streql(XNLineSpace, preedit_attr->name)) {
 	    preedit_attr->value = (void *)malloc(sizeof(long));
-	    *(long*)preedit_attr->value
-		= ic->preedit.line_space;
+	    *(long*)preedit_attr->value = ic->preedit.line_space;
 	    preedit_attr->value_length = sizeof(long);
 	} else if (streql(XNPreeditState, preedit_attr->name)) {
-	    preedit_attr->value = 
-		(void *)malloc(sizeof(XIMPreeditState));
-	    *(XIMPreeditState*)preedit_attr->value
-		= ic->preedit.state;
+	    preedit_attr->value = (void *)malloc(sizeof(XIMPreeditState));
+	    *(XIMPreeditState*)preedit_attr->value = ic->preedit.state;
 	    preedit_attr->value_length = sizeof(XIMPreeditState);
 	} else if (streql(XNFontSet, preedit_attr->name)) {
 	    CARD16 base_len = (CARD16)strlen(ic->preedit.base_font);
@@ -779,8 +832,7 @@ nabi_ic_get_values(NabiIC *ic, IMChangeICStruct *data)
 	    status_attr->value_length = sizeof(XRectangle);
 	} else if (streql(XNAreaNeeded, status_attr->name)) {
 	    status_attr->value = (void *)malloc(sizeof(XRectangle));
-	    *(XRectangle*)status_attr->value
-		= ic->status_attr.area_needed;
+	    *(XRectangle*)status_attr->value = ic->status_attr.area_needed;
 	    status_attr->value_length = sizeof(XRectangle);
 	} else if (streql(XNForeground, status_attr->name)) {
 	    status_attr->value = (void *)malloc(sizeof(long));
@@ -941,6 +993,8 @@ nabi_ic_set_mode(NabiIC *ic, NabiInputMode mode)
     default:
 	break;
     }
+
+    nabi_ic_status_update(ic);
 }
 
 void
@@ -967,7 +1021,13 @@ nabi_ic_preedit_start(NabiIC *ic)
 	preedit_data.todo.return_value = 0;
 	IMCallCallback(nabi_server->xims, (XPointer)&preedit_data);
     } else if (ic->input_style & XIMPreeditPosition) {
-	if (ic->preedit.window == 0)
+	if (ic->preedit.window == NULL)
+	    nabi_ic_preedit_window_new(ic);
+    } else if (ic->input_style & XIMPreeditArea) {
+	if (ic->preedit.window == NULL)
+	    nabi_ic_preedit_window_new(ic);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	if (ic->preedit.window == NULL)
 	    nabi_ic_preedit_window_new(ic);
     }
     ic->preedit.start = True;
@@ -990,6 +1050,10 @@ nabi_ic_preedit_done(NabiIC *ic)
 	IMCallCallback(nabi_server->xims, (XPointer)&preedit_data);
     } else if (ic->input_style & XIMPreeditPosition) {
 	nabi_ic_preedit_hide(ic);
+    } else if (ic->input_style & XIMPreeditArea) {
+	nabi_ic_preedit_hide(ic);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	nabi_ic_preedit_hide(ic);
     }
 
     if (nabi_server->dynamic_event_flow) {
@@ -1008,11 +1072,16 @@ nabi_ic_get_preedit_string(NabiIC *ic, char *utf8, int buflen,
 			   int *len, int *size)
 {
     int i, n;
+    int output_mode = nabi_server->output_mode;
     wchar_t ch;
     wchar_t buf[16];
-    
+
+    if (nabi_server->keyboard_table->output_mode >= 0) {
+	output_mode = nabi_server->keyboard_table->output_mode;
+    }
+
     n = 0;
-    if (nabi_server->output_mode == NABI_OUTPUT_MANUAL) {
+    if (output_mode == NABI_OUTPUT_MANUAL) {
 	/* we use conjoining jamo, U+1100 - U+11FF */
 	if (ic->choseong[0] == 0)
 	    buf[n++] = HCF;
@@ -1030,7 +1099,7 @@ nabi_ic_get_preedit_string(NabiIC *ic, char *utf8, int buflen,
 	    for (i = 0; i <= ic->tindex; i++)
 		buf[n++] = ic->jongseong[i];
 	}
-    } else if (nabi_server->output_mode == NABI_OUTPUT_JAMO) {
+    } else if (output_mode == NABI_OUTPUT_JAMO) {
 	/* we use conjoining jamo, U+1100 - U+11FF */
 	if (ic->choseong[0] == 0)
 	    buf[n++] = HCF;
@@ -1094,19 +1163,31 @@ nabi_ic_get_preedit_string(NabiIC *ic, char *utf8, int buflen,
     *size = hangul_wcharstr_to_utf8str(buf, utf8, buflen);
 }
 
+inline XIMFeedback *
+nabi_ic_preedit_feedback_new(int len, long style)
+{
+    int i;
+    XIMFeedback *feedback = g_new(XIMFeedback, len);
+
+    for (i = 0; i < len; ++i)
+	feedback[i] = style;
+
+    return feedback;
+}
+
 void
 nabi_ic_preedit_insert(NabiIC *ic)
 {
     int len, size;
     char buf[48] = { 0, };
-    char *compound_text;
-    IMPreeditCBStruct data;
-    XIMText text;
-    XIMFeedback feedback[4] = { XIMReverse, 0, 0, 0 };
 
     nabi_ic_get_preedit_string(ic, buf, sizeof(buf), &len, &size);
 
     if (ic->input_style & XIMPreeditCallbacks) {
+	char *compound_text;
+	IMPreeditCBStruct data;
+	XIMText text;
+
 	compound_text = utf8_to_compound_text(buf);
 
 	data.major_code = XIM_PREEDIT_DRAW;
@@ -1118,16 +1199,23 @@ nabi_ic_preedit_insert(NabiIC *ic)
 	data.todo.draw.chg_length = 0;
 	data.todo.draw.text = &text;
 
-	text.feedback = feedback;
+	text.feedback = nabi_ic_preedit_feedback_new(len, XIMReverse);
 	text.encoding_is_wchar = False;
 	text.string.multi_byte = compound_text;
 	text.length = strlen(compound_text);
 
 	IMCallCallback(nabi_server->xims, (XPointer)&data);
 	XFree(compound_text);
+	g_free(text.feedback);
     } else if (ic->input_style & XIMPreeditPosition) {
-	nabi_ic_preedit_draw_string(ic, buf, size);
 	nabi_ic_preedit_show(ic);
+	nabi_ic_preedit_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditArea) {
+	nabi_ic_preedit_show(ic);
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	nabi_ic_preedit_show(ic);
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
     }
     ic->preedit.prev_length = len;
 }
@@ -1137,10 +1225,6 @@ nabi_ic_preedit_update(NabiIC *ic)
 {
     int len, size;
     char buf[48] = { 0, };
-    char *compound_text;
-    IMPreeditCBStruct data;
-    XIMText text;
-    XIMFeedback feedback[4] = { XIMReverse, 0, 0, 0 };
 
     nabi_ic_get_preedit_string(ic, buf, sizeof(buf), &len, &size);
 
@@ -1150,6 +1234,10 @@ nabi_ic_preedit_update(NabiIC *ic)
     }
 
     if (ic->input_style & XIMPreeditCallbacks) {
+	char *compound_text;
+	XIMText text;
+	IMPreeditCBStruct data;
+
 	compound_text = utf8_to_compound_text(buf);
 
 	data.major_code = XIM_PREEDIT_DRAW;
@@ -1161,7 +1249,7 @@ nabi_ic_preedit_update(NabiIC *ic)
 	data.todo.draw.chg_length = ic->preedit.prev_length;
 	data.todo.draw.text = &text;
 
-	text.feedback = feedback;
+	text.feedback = nabi_ic_preedit_feedback_new(len, XIMReverse);
 	text.encoding_is_wchar = False;
 	text.string.multi_byte = compound_text;
 	text.length = strlen(compound_text);
@@ -1170,6 +1258,10 @@ nabi_ic_preedit_update(NabiIC *ic)
 	XFree(compound_text);
     } else if (ic->input_style & XIMPreeditPosition) {
 	nabi_ic_preedit_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditArea) {
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	nabi_ic_preedit_gdk_draw_string(ic, buf, size);
     }
     ic->preedit.prev_length = len;
 }
@@ -1198,7 +1290,7 @@ nabi_ic_preedit_clear(NabiIC *ic)
 
 	IMCallCallback(nabi_server->xims, (XPointer)&data);
     } else if (ic->input_style & XIMPreeditPosition) {
-	if (ic->preedit.window == 0)
+	if (ic->preedit.window == NULL)
 	    return;
 
 	/* we resize the preedit window instead of unmap
@@ -1209,82 +1301,22 @@ nabi_ic_preedit_clear(NabiIC *ic)
 	XResizeWindow(nabi_server->display, ic->preedit.window,
 		      ic->preedit.width, ic->preedit.height);
 	 */
-	XUnmapWindow(nabi_server->display, ic->preedit.window);
+	gdk_window_hide(ic->preedit.window);
+    } else if (ic->input_style & XIMPreeditArea) {
+	if (ic->preedit.window != NULL)
+	    gdk_window_hide(ic->preedit.window);
+    } else if (ic->input_style & XIMPreeditNothing) {
+	if (ic->preedit.window != NULL)
+	    gdk_window_hide(ic->preedit.window);
     }
     ic->preedit.prev_length = 0;
 }
 
-Bool
-nabi_ic_commit(NabiIC *ic)
+static void
+nabi_ic_commit_utf8(NabiIC *ic, char *utf8_str)
 {
-    int i, n;
     IMCommitStruct commit_data;
-    wchar_t buf[16];
-    char utf8buf[48];
     char *compound_text;
- 
-    buf[0] = L'\0';
-
-    if (nabi_ic_is_empty(ic))
-	return False;
-
-    n = 0;
-    if (nabi_server->output_mode == NABI_OUTPUT_MANUAL) {
-	/* we use conjoining jamo, U+1100 - U+11FF */
-	if (ic->choseong[0] == 0)
-	    buf[n++] = HCF;
-	else {
-	    for (i = 0; i <= ic->lindex; i++)
-		buf[n++] = ic->choseong[i];
-	}
-	if (ic->jungseong[0] == 0)
-	    buf[n++] = HJF;
-	else {
-	    for (i = 0; i <= ic->vindex; i++)
-		buf[n++] = ic->jungseong[i];
-	}
-	if (ic->jongseong[0] != 0) {
-	    for (i = 0; i <= ic->tindex; i++)
-		buf[n++] = ic->jongseong[i];
-	}
-    } else if (nabi_server->output_mode == NABI_OUTPUT_JAMO) {
-	/* we use conjoining jamo, U+1100 - U+11FF */
-	if (ic->choseong[0] == 0)
-	    buf[n++] = HCF;
-	else
-	    buf[n++] = ic->choseong[0];
-	if (ic->jungseong[0] == 0)
-	    buf[n++] = HJF;
-	else
-	    buf[n++] = ic->jungseong[0];
-	if (ic->jongseong[0] != 0)
-	    buf[n++] = ic->jongseong[0];
-    } else {
-	/* use hangul syllables (U+AC00 - U+D7AF)
-	 * and compatibility jamo (U+3130 - U+318F) */
-	wchar_t ch;
-	ch = hangul_jamo_to_syllable(ic->choseong[0],
-			 ic->jungseong[0],
-			 ic->jongseong[0]);
-	if (ch)
-	    buf[n++] = ch;
-	else {
-	    if (ic->choseong[0]) {
-		ch = hangul_choseong_to_cjamo(ic->choseong[0]);
-		buf[n++] = ch;
-	    }
-	    if (ic->jungseong[0]) {
-		ch = hangul_jungseong_to_cjamo(ic->jungseong[0]);
-		buf[n++] = ch;
-	    }
-	    if (ic->jongseong[0]) {
-		ch = hangul_jongseong_to_cjamo(ic->jongseong[0]);
-		buf[n++] = ch;
-	    }
-	}
-    }
-    buf[n] = L'\0';
-    nabi_ic_buf_clear(ic);
 
     /* According to XIM Spec, We should delete preedit string here 
      * befor commiting the string. but it makes too many flickering
@@ -1296,8 +1328,10 @@ nabi_ic_commit(NabiIC *ic)
     if (!(ic->input_style & XIMPreeditPosition))
 	nabi_ic_preedit_clear(ic);
 
-    hangul_wcharstr_to_utf8str(buf, utf8buf, sizeof(utf8buf));
-    compound_text = utf8_to_compound_text(utf8buf);
+    compound_text = utf8_to_compound_text(utf8_str);
+
+    commit_data.major_code = XIM_COMMIT;
+    commit_data.minor_code = 0;
     commit_data.connect_id = ic->connect_id;
     commit_data.icid = ic->id;
     commit_data.flag = XimLookupChars;
@@ -1309,31 +1343,36 @@ nabi_ic_commit(NabiIC *ic)
     /* we delete preedit string here when PreeditPosition */
     if (ic->input_style & XIMPreeditPosition)
 	nabi_ic_preedit_clear(ic);
-
-    return True;
 }
 
 Bool
+nabi_ic_commit(NabiIC *ic)
+{
+    int len, size;
+    char buf[64];
+ 
+    if (nabi_ic_is_empty(ic))
+	return False;
+
+    nabi_ic_get_preedit_string(ic, buf, sizeof(buf), &len, &size);
+    nabi_ic_buf_clear(ic);
+
+    nabi_ic_commit_utf8(ic, buf);
+    return True;
+}
+
+static Bool
 nabi_ic_commit_unicode(NabiIC *ic, wchar_t ch)
 {
-    IMCommitStruct commit_data;
-    wchar_t buf[16];
-    char utf8buf[48];
-    char *compound_text;
+    int size = 0;
+    char buf[64];
  
-    buf[0] = ch;
-    buf[1] = L'\0';
+    size += hangul_wchar_to_utf8(ch, buf + size, sizeof(buf) - size);
+    buf[size] = '\0';
 
-    hangul_wcharstr_to_utf8str(buf, utf8buf, sizeof(utf8buf));
-    compound_text = utf8_to_compound_text(utf8buf);
-    commit_data.connect_id = ic->connect_id;
-    commit_data.icid = ic->id;
-    commit_data.flag = XimLookupChars;
-    commit_data.commit_string = compound_text;
-
-    IMCommitString(nabi_server->xims, (XPointer)&commit_data);
-    XFree(compound_text);
+    nabi_ic_buf_clear(ic);
 
+    nabi_ic_commit_utf8(ic, buf);
     return True;
 }
 
@@ -1344,6 +1383,15 @@ nabi_ic_commit_keyval(NabiIC *ic, wchar_t ch, KeySym keyval)
     if (ch != 0)
 	return nabi_ic_commit_unicode(ic, ch);
 
+    /* we forward all keys which does not accepted by nabi */
+    return False;
+#if 0
+    /* This code was for mozilla bug.
+     * Mozilla crashed when xim forward key event on hangul mode,
+     * so nabi commited the key value instead of forwaring.
+     * But now it seems that this bug is fixed, so every key event
+     * will be forwarded */
+
     /* ISO10646 charcode keyval */
     if ((keyval & 0xff000000) == 0x01000000)
 	return nabi_ic_commit_unicode(ic, keyval & 0x00ffffff);
@@ -1354,6 +1402,112 @@ nabi_ic_commit_keyval(NabiIC *ic, wchar_t ch, KeySym keyval)
 
     /* Forward all other keys */
     return False;
+#endif
+}
+
+void
+nabi_ic_status_start(NabiIC *ic)
+{
+    if (!nabi_server->show_status)
+	return;
+
+    if (ic->input_style & XIMStatusCallbacks) {
+	IMStatusCBStruct data;
+	char *compound_text;
+	XIMText text;
+	XIMFeedback feedback[4] = { 0, 0, 0, 0 };
+
+	compound_text = "";
+
+	data.major_code = XIM_STATUS_START;
+	data.minor_code = 0;
+	data.connect_id = ic->connect_id;
+	data.icid = ic->id;
+	data.todo.draw.data.text = &text;
+	data.todo.draw.type = XIMTextType;
+
+	text.feedback = feedback;
+	text.encoding_is_wchar = False;
+	text.string.multi_byte = compound_text;
+	text.length = strlen(compound_text);
+
+	IMCallCallback(nabi_server->xims, (XPointer)&data);
+    }
+    g_print("Status start\n");
+}
+
+void
+nabi_ic_status_done(NabiIC *ic)
+{
+    if (!nabi_server->show_status)
+	return;
+
+    if (ic->input_style & XIMStatusCallbacks) {
+	IMStatusCBStruct data;
+	char *compound_text;
+	XIMText text;
+	XIMFeedback feedback[4] = { 0, 0, 0, 0 };
+
+	compound_text = "";
+
+	data.major_code = XIM_STATUS_DONE;
+	data.minor_code = 0;
+	data.connect_id = ic->connect_id;
+	data.icid = ic->id;
+	data.todo.draw.data.text = &text;
+	data.todo.draw.type = XIMTextType;
+
+	text.feedback = feedback;
+	text.encoding_is_wchar = False;
+	text.string.multi_byte = compound_text;
+	text.length = strlen(compound_text);
+
+	IMCallCallback(nabi_server->xims, (XPointer)&data);
+    }
+    g_print("Status done\n");
+}
+
+void
+nabi_ic_status_update(NabiIC *ic)
+{
+    if (!nabi_server->show_status)
+	return;
+
+    if (ic->input_style & XIMStatusCallbacks) {
+	IMStatusCBStruct data;
+	char *status_str;
+	char *compound_text;
+	XIMText text;
+	XIMFeedback feedback[4] = { 0, 0, 0, 0 };
+
+	switch (ic->mode) {
+	case NABI_INPUT_MODE_DIRECT:
+	    status_str = "영어";
+	    break;
+	case NABI_INPUT_MODE_COMPOSE:
+	    status_str = "한글";
+	    break;
+	default:
+	    status_str = "";
+	    break;
+	}
+	compound_text = utf8_to_compound_text(status_str);
+
+	data.major_code = XIM_STATUS_DRAW;
+	data.minor_code = 0;
+	data.connect_id = ic->connect_id;
+	data.icid = ic->id;
+	data.todo.draw.data.text = &text;
+	data.todo.draw.type = XIMTextType;
+
+	text.feedback = feedback;
+	text.encoding_is_wchar = False;
+	text.string.multi_byte = compound_text;
+	text.length = strlen(compound_text);
+
+	IMCallCallback(nabi_server->xims, (XPointer)&data);
+    }
+    g_print("Status draw\n");
 }
 
 static int
@@ -1430,7 +1584,7 @@ nabi_ic_popup_candidate_window (NabiIC *ic)
 	if (index >= 0) {
 	    ptr = (const NabiCandidateItem **)
 		    nabi_server->candidate_table[index] + 1;
-	    ic->candidate = nabi_candidate_new(NULL, 10, ptr, parent,
+	    ic->candidate = nabi_candidate_new(NULL, 9, ptr, parent,
 				    &nabi_ic_candidate_commit_cb, ic);
 	    return True;
 	}
