@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003 Choe Hwanjin
+ * Copyright (C) 2003,2004 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -47,7 +47,7 @@ struct _NabiConnect {
 };
 
 struct _PreeditAttributes {
-    Window          window;         /* where to draw the preedit string */
+    GdkWindow*      window;         /* where to draw the preedit string */
     int             width;          /* preedit area width */
     int             height;         /* preedit area height */
     XPoint          spot;           /* window position */
@@ -56,7 +56,7 @@ struct _PreeditAttributes {
     XRectangle      area_needed;    /* area needed */
 
     Colormap        cmap;           /* colormap */
-    GC              gc;             /* gc */
+    GdkGC*          gc;             /* gc */
     unsigned long   foreground;     /* foreground */
     unsigned long   background;     /* background */
 
@@ -148,6 +148,10 @@ void    nabi_ic_preedit_insert(NabiIC *ic);
 void    nabi_ic_preedit_update(NabiIC *ic);
 void    nabi_ic_preedit_clear(NabiIC *ic);
 
+void    nabi_ic_status_start(NabiIC *ic);
+void    nabi_ic_status_done(NabiIC *ic);
+void    nabi_ic_status_update(NabiIC *ic);
+
 Bool    nabi_ic_commit(NabiIC *ic);
 Bool    nabi_ic_commit_keyval(NabiIC *ic, wchar_t ch, KeySym keyval);
 
