@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003 Choe Hwanjin
+ * Copyright (C) 2003,2004 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,24 +19,11 @@
 #ifndef __NABI_H_
 #define __NABI_H_
 
-typedef struct _NabiKeyboardMap NabiKeyboardMap;
-typedef struct _NabiComposeMap NabiComposeMap;
+#include <gtk/gtk.h>
+#include <X11/Xlib.h>
+
 typedef struct _NabiApplication NabiApplication;
 
-#define KEYBOARD_MAP_SIZE 94
-struct _NabiKeyboardMap {
-    gchar*           filename;
-    gchar*           name;
-    gint             type;
-    wchar_t          map[KEYBOARD_MAP_SIZE];
-};
-
-struct _NabiComposeMap {
-    gchar*                  name;
-    NabiComposeItem**       map;
-    gint                    size;
-};
-
 struct _NabiApplication {
     gchar*	    xim_name;		/* xim server name from config file */
     gchar*	    optional_xim_name;	/* xim server name from command line */
@@ -48,13 +35,16 @@ struct _NabiApplication {
     gchar*	    session_id;
 
     gchar*          theme;
-    gchar*          keyboard_map_filename;
-    gchar*          compose_map_filename;
+    int             icon_size;
+
+    gchar*          keyboard_table_name;
+    gchar*          keyboard_table_dir;
+    gchar*          compose_table_name;
+    gchar*          compose_table_dir;
     gchar*	    candidate_table_filename;
 
-    GSList 	    *keyboard_maps;
-
-    NabiComposeMap  compose_map;
+    gchar*          trigger_keys;
+    gchar*          candidate_keys;
 
     /* xim server option */
     gboolean        dvorak;
@@ -82,9 +72,14 @@ void nabi_app_setup_server(void);
 void nabi_app_quit(void);
 void nabi_app_free(void);
 void nabi_save_config_file(void);
+void nabi_app_set_theme(const char *name);
+void nabi_app_set_dvorak(gboolean state);
+void nabi_app_set_keyboard(const char *name);
+void nabi_app_set_candidate_font(const char *font);
+void nabi_app_set_trigger_keys(char **keys);
+void nabi_app_set_candidate_keys(char **keys);
 
 GtkWidget* nabi_app_create_main_widget(void);
-GtkWidget* nabi_create_hanja_window(NabiIC *ic, const wchar_t* ch);
 
 #endif /* __NABI_H_ */
 /* vim: set ts=8 sw=4 : */
